import re


class _string:
    @staticmethod
    def camel_to_snake(string: str) -> str:
        """
        >>> _string.camel_to_snake('CamelCaseString')
        'camel_case_string'

        Args:
            string:

        Returns:

        """
        string = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", string)
        string = re.sub("([a-z0-9])([A-Z])", r"\1_\2", string).lower()
        return string

    @staticmethod
    def snake_to_camel(string: str) -> str:
        """
        >>> _string.snake_to_camel('snake_case_string')
        'SnakeCaseString'

        Args:
            string:

        Returns:

        """
        return "".join(i.capitalize() for i in string.split("_"))


__all__ = [
    "_string"
]



from typing import Any

import sqlparse


class _sql:
    @staticmethod
    def format(sql: str, **kwargs: Any) -> str:
        if not kwargs:
            kwargs = dict(
                reindent=True,
                keyword_case="upper",
                identifier_case="lower",
                strip_comments=True
            )
        sql = sqlparse.format(sql, **kwargs)
        return sql


__all__ = [
    "_sql"
]



import math


class _math:
    @staticmethod
    def int_ceil(num: float) -> int:
        """
        >>> _math.int_ceil(10 / 3)
        4

        Args:
            num:

        Returns:

        """
        return math.ceil(num)

    @staticmethod
    def decimal_ceil(
            num: float,
            places: int = 2
    ) -> float:
        """
        >>> _math.decimal_ceil(3.14159)
        3.15

        Args:
            num:
            places:

        Returns:

        """
        if places >= 0:
            scale = 10.0 ** places
            return math.ceil(num * scale) / scale
        else:
            raise ValueError(
                f"Invalid value for 'places': "
                f"Expected `int` and >= 0, "
                f"but got value: {places!r}"
            )

    @staticmethod
    def format_number(
            number: int | float | str,
            places: int = 2
    ) -> str:
        """
        >>> _math.format_number(3.1)
        '3.10'

        Args:
            number:
            places:

        Returns:

        """
        integer_str, decimal_str = str(float(number)).split(".")
        decimal_str = (decimal_str + "0" * (places - len(decimal_str)))[:places]
        number_str = ".".join([integer_str, decimal_str])
        return number_str


__all__ = [
    "_math"
]



import os
import platform
import subprocess
import zipfile
from typing import Literal, cast

from utils._assets import _assets


class _file:
    COMPRESS_TYPE = Literal["zip", "rar"]
    DECOMPRESS_TYPE = Literal["zip", "rar"]

    @staticmethod
    def compress(
            dir_path: str,
            compress_type: COMPRESS_TYPE | None = None
    ) -> str | None:
        dir_path = os.path.abspath(dir_path)
        if not os.path.isdir(dir_path):
            return None

        if compress_type is None:
            compress_type = "zip"

        compress_file_path = os.path.join(
            os.path.dirname(dir_path), os.path.basename(dir_path) + "." + compress_type
        )

        if compress_type == "zip":
            with zipfile.ZipFile(compress_file_path, "w", zipfile.ZIP_DEFLATED) as zf:
                for root, dirs, files in os.walk(dir_path):
                    for file in files:
                        file_path = os.path.join(root, file)
                        if os.path.abspath(file_path) == compress_file_path:
                            continue
                        rel_file_path = os.path.relpath(file_path, dir_path)
                        zf.write(file_path, arcname=rel_file_path)
            return compress_file_path

        if compress_type == "rar":
            args = [
                _assets.get_assets_file_path("Rar.exe"),
                "a",
                "-r",
                "-inul",
                "-ep1",
                compress_file_path,
                os.path.join(dir_path, "*")
            ]
            if subprocess.run(args).returncode == 0:
                return compress_file_path

        return None

    @staticmethod
    def decompress(
            compress_file_path: str,
            dir_path: str | None = None
    ) -> str | None:
        compress_file_path = os.path.abspath(compress_file_path)
        if not os.path.isfile(compress_file_path):
            return None

        if dir_path is None:
            dir_path = os.path.splitext(compress_file_path)[0]
        else:
            dir_path = os.path.abspath(dir_path)

        os.makedirs(dir_path, exist_ok=True)

        with open(compress_file_path, "rb") as file:
            data = file.read(8)
        text = data.hex()
        if text == "504b030414000000":
            decompress_type = "zip"
        elif text == "526172211a070100":
            decompress_type = "rar"
        else:
            return None

        if decompress_type == "zip":
            with zipfile.ZipFile(compress_file_path, "r") as zf:
                zf.extractall(dir_path)
            return dir_path

        if decompress_type == "rar":
            args = [
                _assets.get_assets_file_path("UnRAR.exe"),
                "x",
                "-o+",
                "-inul",
                compress_file_path,
                dir_path
            ]
            if subprocess.run(args).returncode == 0:
                return dir_path

        return None

    @staticmethod
    def get_file_paths_and_dir_paths(path: str) -> tuple[list[str], list[str]]:
        file_paths = []
        dir_paths = []

        path = os.path.abspath(path)

        with os.scandir(path) as entries:
            for entry in entries:
                System = Literal["Windows", "Linux", "Darwin"]
                system: System = cast(System, platform.system())
                if system == "Windows":
                    from nt import DirEntry
                elif system == "Linux":
                    from posix import DirEntry
                elif system == "Darwin":
                    from posix import DirEntry
                else:
                    raise TypeError(
                        f"Invalid type for 'system': "
                        f"Expected `Literal[\"Windows\",\"Linux\",\"Darwin\"]`, "
                        f"but got {type(system).__name__!r} (value: {system!r})"
                    )

                entry: DirEntry
                if entry.is_file():
                    file_paths.append(entry.path)
                elif entry.is_dir():
                    dir_paths.append(entry.path)
                    sub_file_paths, sub_dir_paths = _file.get_file_paths_and_dir_paths(entry.path)
                    file_paths.extend(sub_file_paths)
                    dir_paths.extend(sub_dir_paths)

        return file_paths, dir_paths


__all__ = [
    "_file"
]



import pandas as pd


class _pandas:
    @staticmethod
    def split(
            df: pd.DataFrame,
            num_parts: int | None = None,
            part_size: int | None = None
    ) -> list[pd.DataFrame]:
        if df.empty:
            return []

        if num_parts:
            avg_size = len(df) // num_parts
            remainder = len(df) % num_parts

            result = []
            start = 0
            for i in range(num_parts):
                end = start + avg_size + (1 if i < remainder else 0)
                result.append(df.iloc[start:end])
                start = end
            return result

        elif part_size:
            return [df.iloc[i:i + part_size] for i in range(0, len(df), part_size)]

        else:
            raise ValueError(
                f"Either num_parts: {num_parts!r} or part_size: {part_size!r} must be provided"
            )


__all__ = [
    "_pandas"
]



import os
from pathlib import Path
from tempfile import NamedTemporaryFile

import pillow_avif  # type: ignore
from PIL import Image


class _image:
    @staticmethod
    def to_jpg(
            image_file_path: str,
            /,
            jpg_file_path: str | None = None,
            quality: int = 100,
            keep_original: bool = False
    ) -> str | None:
        try:
            if jpg_file_path is None:
                image_file_prefix, _ = os.path.splitext(image_file_path)
                jpg_file_path = image_file_prefix + ".jpg"

            if image_file_path == jpg_file_path:
                return jpg_file_path

            jpg_dir_path = os.path.dirname(jpg_file_path)
            if not os.path.exists(jpg_dir_path):
                os.makedirs(jpg_dir_path, exist_ok=True)

            with Image.open(image_file_path) as image:
                if image.mode in ("RGBA", "LA"):
                    background = Image.new("RGB", image.size, (255, 255, 255))
                    background.paste(image, mask=image.split()[-1])
                    image = background
                elif image.mode == "P":
                    image.seek(0)  # image.n_frames
                    image = image.convert("RGB")
                elif image.mode == "RGB":
                    pass
                else:
                    return None

                with NamedTemporaryFile(suffix=".jpg", delete=False, dir=jpg_dir_path) as ntf:
                    temp_file_path = ntf.name
                    image.save(temp_file_path, "JPEG", quality=quality)
                os.replace(temp_file_path, jpg_file_path)

            if not keep_original:
                p = Path(image_file_path)
                if p.exists() and str(p.resolve()) == str(p.absolute()):
                    os.remove(image_file_path)

            return jpg_file_path
        except Exception as e:  # noqa
            return None


__all__ = [
    "_image"
]



import os
import re
import sys
from typing import Any, Final
from urllib import parse

import httpx
import tldextract
from furl import furl
from typing_extensions import Literal
from w3lib.url import canonicalize_url

from utils._headers import _headers


# noinspection PyShadowingNames
class _url:
    @staticmethod
    def get_furl_obj(url: str) -> furl:
        return furl(url)

    @staticmethod
    def get_parse_result(url: str) -> parse.ParseResult:
        parse_result = parse.urlparse(url)
        return parse_result

    @staticmethod
    def get_origin_path(url: str) -> str:
        """
        >>> _url.get_origin_path("https://github.com/search?q=owner%3Amathewgeola+cwb-utils&type=repositories")
        'https://github.com/search'

        Args:
            url:

        Returns:

        """
        furl_obj = _url.get_furl_obj(url)
        origin_path = str(furl_obj.origin) + str(furl_obj.path)
        return origin_path

    @staticmethod
    def is_valid(url: str) -> bool:
        """
        >>> _url.is_valid("https://www.baidu.com/")
        True

        Args:
            url:

        Returns:

        """
        try:
            parse_result = _url.get_parse_result(url)
            scheme, netloc = parse_result.scheme, parse_result.netloc
            if not scheme:
                return False
            if not netloc:
                return False
            if scheme not in ("http", "https"):
                return False
            return True
        except ValueError:
            return False

    @staticmethod
    def quote(
            url: str,
            safe: str | None = None,
            encoding: str = "utf-8",
            quote_type: Literal["encodeURI", "encodeURIComponent", "browser"] | None = "browser"
    ) -> str:
        """
        >>> _url.quote("https://www.baidu.com/s?wd=你好")
        'https://www.baidu.com/s?wd=%E4%BD%A0%E5%A5%BD'

        Args:
            url:
            safe:
            encoding:
            quote_type:

        Returns:

        """
        if quote_type == "encodeURI":
            safe = ";/?:@&=+$,-_.!~*'()#"
        elif quote_type == "encodeURIComponent":
            safe = "-_.~"
        elif quote_type == "browser":
            safe = ";/?:@&=+$,-_.!~*'()"
            parsed = parse.urlparse(url)
            path = parse.quote(parsed.path, safe=safe)
            query_pairs = parse.parse_qsl(parsed.query, keep_blank_values=True)
            encoded_query = "&".join(
                f"{k}={parse.quote(v, safe='-_.~', encoding=encoding)}"
                for k, v in query_pairs
            )
            fragment = parse.quote(parsed.fragment, safe='-_.~', encoding=encoding)
            return parse.urlunparse((
                parsed.scheme,
                parsed.netloc,
                path,
                parsed.params,
                encoded_query,
                fragment
            ))
        else:
            if safe is None:
                safe = "/"

        return parse.quote(url, safe=safe, encoding=encoding)

    @staticmethod
    def unquote(url: str, encoding: str = "utf-8") -> str:
        """
        >>> _url.unquote("https://www.baidu.com/s?wd=%E4%BD%A0%E5%A5%BD")
        'https://www.baidu.com/s?wd=你好'

        Args:
            url:
            encoding:

        Returns:

        """
        return parse.unquote(url, encoding=encoding)

    @staticmethod
    def encode(params: dict[str, Any]) -> str:
        """
        >>> _url.encode({"a": "1", "b": "2"})
        'a=1&b=2'

        Args:
            params:

        Returns:

        """
        return parse.urlencode(params)

    @staticmethod
    def decode(url: str) -> dict[str, str]:
        """
        >>> _url.decode("xxx?a=1&b=2")
        {'a': '1', 'b': '2'}

        Args:
            url:

        Returns:

        """
        params = dict()

        lst = url.split("?", maxsplit=1)[-1].split("&")
        for i in lst:
            key, value = i.split("=", maxsplit=1)
            params[key] = _url.unquote(value)

        return params

    @staticmethod
    def join_url(base_url: str, url: str) -> str:
        """
        >>> _url.join_url("https://www.baidu.com/", "/s?ie=UTF-8&wd=cwb-utils")
        'https://www.baidu.com/s?ie=UTF-8&wd=cwb-utils'

        Args:
            base_url:
            url:

        Returns:

        """
        return parse.urljoin(base_url, url)

    @staticmethod
    def join_params(url: str, params: dict[str, Any]) -> str:
        """
        >>> _url.join_params("https://www.baidu.com/s", {"wd": "你好"})
        'https://www.baidu.com/s?wd=%E4%BD%A0%E5%A5%BD'

        Args:
            url:
            params:

        Returns:

        """
        if not params:
            return url

        params = _url.encode(params)
        separator = "?" if "?" not in url else "&"
        return url + separator + params

    @staticmethod
    def get_params(url: str) -> dict[str, str]:
        """
        >>> _url.get_params("https://www.baidu.com/s?wd=cwb-utils")
        {'wd': 'cwb-utils'}

        Args:
            url:

        Returns:

        """
        furl_obj = _url.get_furl_obj(url)
        params = dict(furl_obj.query.params)
        return params

    @staticmethod
    def get_param(url: str, key: str, default: Any | None = None) -> Any:
        """
        >>> _url.get_param("https://www.baidu.com/s?wd=cwb-utils", "wd")
        'cwb-utils'

        Args:
            url:
            key:
            default:

        Returns:

        """
        params = _url.get_params(url)
        param = params.get(key, default)
        return param

    @staticmethod
    def get_url_params(url: str) -> tuple[str, dict[str, str]]:
        """
        >>> _url.get_url_params("https://www.baidu.com/s?wd=cwb-utils")
        ('https://www.baidu.com/s', {'wd': 'cwb-utils'})

        Args:
            url:

        Returns:

        """
        root_url = ""
        params = dict()

        if "?" in url:
            root_url = url.split("?", maxsplit=1)[0]
            params = _url.get_params(url)
        else:
            if re.search("[&=]", url) and not re.search("/", url):
                params = _url.get_params(url)
            else:
                root_url = url

        return root_url, params

    @staticmethod
    def get_domain(url: str) -> str:
        """
        >>> _url.get_domain("https://image.baidu.com/search/index?word=cwb-utils")
        'baidu'

        Args:
            url:

        Returns:

        """
        er = tldextract.extract(url)
        domain = er.domain
        return domain

    @staticmethod
    def get_subdomain(url: str) -> str:
        """
        >>> _url.get_subdomain("https://image.baidu.com/search/index?word=cwb-utils")
        'image'

        Args:
            url:

        Returns:

        """
        er = tldextract.extract(url)
        subdomain = er.subdomain
        return subdomain

    @staticmethod
    def canonicalize(url: str) -> str:
        """
        >>> _url.canonicalize("https://www.baidu.com/s?wd=cwb-utils")
        'https://www.baidu.com/s?wd=cwb-utils'

        Args:
            url:

        Returns:

        """
        return canonicalize_url(url)

    @staticmethod
    def to_file_path(
            url: str,
            headers: dict[str, str] | None = None,
            file_path: str | None = None,
            dir_path: str | None = None,
            file_name: str | None = None,
            file_prefix: str | None = None,
            file_suffix: str | None = None,
            use_cache: bool = True,
            chunk_size: int = 64 * 1024
    ) -> str | None:
        if not _url.is_valid(url):
            return None

        if headers is None:
            headers = _headers.get_default()

        if file_path is None:
            if dir_path is None:
                sys_argv0 = sys.argv[0]
                if not os.path.isfile(sys_argv0):
                    return None
                dir_path = os.path.dirname(sys_argv0)
            if file_name is None:
                if file_prefix is not None and file_suffix is not None:
                    file_name = file_prefix + file_suffix
                else:
                    if file_prefix is None:
                        file_prefix, _ = os.path.splitext(_url.get_furl_obj(url).path.segments[-1])
                    if file_suffix is None:
                        _, file_suffix = os.path.splitext(_url.get_furl_obj(url).path.segments[-1])
                        if file_suffix is None or file_suffix == "":
                            response = httpx.head(url, headers=headers)
                            if (content_type := response.headers.get("content-type")) is not None:
                                # todo：Add more content_type.
                                content_type_to_file_suffix: Final[dict[str, str]] = {
                                    "image/png": "png",
                                    "image/gif": "gif",
                                    "text/html;charset=utf-8": "html",
                                    "text/javascript; charset=utf-8": "js",
                                }
                                file_ext = content_type_to_file_suffix[content_type]
                                file_suffix = "." + file_ext
                    file_name = file_prefix + file_suffix
            file_path = os.path.join(dir_path, file_name)

        file_path = os.path.abspath(file_path)

        if use_cache:
            if os.path.exists(file_path):
                return file_path

        dir_path = os.path.dirname(file_path)
        if not os.path.exists(dir_path):
            os.makedirs(dir_path)

        try:
            with httpx.Client(timeout=None, follow_redirects=True) as client:
                with client.stream("GET", url, headers=headers) as response:
                    response.raise_for_status()
                    with open(file_path, "wb") as file:
                        for chunk in response.iter_bytes(chunk_size=chunk_size):
                            file.write(chunk)
        except Exception as e:  # noqa
            file_path = None

        return file_path


__all__ = [
    "_url"
]



from __future__ import annotations

import logging
import os
import sys
from logging import handlers
from typing import Any, Literal, Final, cast

import colorlog


class _logger:
    FRAMEWORK: Final[int] = 55
    SUCCESS: Final[int] = 25

    logging.addLevelName(FRAMEWORK, "FRAMEWORK")
    logging.addLevelName(SUCCESS, "SUCCESS")

    LoggerLevel = Literal["DEBUG", "INFO", "SUCCESS", "WARNING", "ERROR", "CRITICAL", "FRAMEWORK"]

    class Logger(logging.Logger):
        def framework(
                self,
                msg: str,
                *args: Any,
                level: _logger.LoggerLevel = "DEBUG",
                **kwargs: Any
        ) -> None:
            if self.isEnabledFor(_logger.FRAMEWORK):
                msg = f"[{level}] {msg}"
                kwargs.setdefault("stacklevel", 2)
                self._log(_logger.FRAMEWORK, msg, args, **kwargs)

        def success(self, msg: str, *args: Any, **kwargs: Any) -> None:
            if self.isEnabledFor(_logger.SUCCESS):
                kwargs.setdefault("stacklevel", 2)
                self._log(_logger.SUCCESS, msg, args, **kwargs)

    logging.setLoggerClass(Logger)

    class RotatingFileHandler(handlers.RotatingFileHandler):
        def __init__(self, *args: Any, **kwargs: Any) -> None:
            super().__init__(*args, **kwargs)
            self._zero_padding: Final[int] = len(str(self.backupCount)) if self.backupCount > 0 else 0

        def doRollover(self) -> None:
            if self.stream:
                self.stream.close()
                self.stream = None  # noqa

            if self.backupCount > 0:
                oldest = self._format_backup_filename(self.backupCount)
                if os.path.exists(oldest):
                    os.remove(oldest)

                for i in range(self.backupCount - 1, 0, -1):
                    sfn = self._format_backup_filename(i)
                    dfn = self._format_backup_filename(i + 1)
                    if os.path.exists(sfn):
                        os.rename(sfn, dfn)

                dfn = self._format_backup_filename(1)
                if os.path.exists(self.baseFilename):
                    os.rename(self.baseFilename, dfn)

            self.mode = "w"
            self.stream = self._open()

        def _format_backup_filename(self, index: int) -> str:
            base, ext = os.path.splitext(self.baseFilename)
            padded = str(index).zfill(self._zero_padding)
            return f"{base}.{padded}{ext}"

    @staticmethod
    def _set_console_handler(
            logger: Logger,
            console_level: str | int, console_fmt: str
    ) -> Logger:
        console_handler = logging.StreamHandler()
        console_handler.setLevel(console_level)
        green = {
            "DEBUG": "green",
            "INFO": "green",
            "SUCCESS": "green",
            "WARNING": "green",
            "ERROR": "green",
            "CRITICAL": "green",
            "FRAMEWORK": "green",
        }
        bold_cyan = {
            "DEBUG": "bold_cyan",
            "INFO": "bold_cyan",
            "SUCCESS": "bold_cyan",
            "WARNING": "bold_cyan",
            "ERROR": "bold_cyan",
            "CRITICAL": "bold_cyan",
            "FRAMEWORK": "bold_cyan",
        }
        log_colors = {
            "DEBUG": "light_blue",
            "INFO": "light_white",
            "SUCCESS": "light_green",
            "WARNING": "light_yellow",
            "ERROR": "light_red",
            "CRITICAL": "bg_red,light_white",
            "FRAMEWORK": "light_red,bold",
        }
        secondary_log_colors = dict(
            asctime=green,
            name=bold_cyan,
            levelname=log_colors,
            process=bold_cyan,
            processName=bold_cyan,
            thread=bold_cyan,
            threadName=bold_cyan,
            pathname=bold_cyan,
            funcName=bold_cyan,
            lineno=bold_cyan,
            message=log_colors,
        )
        console_formatter = colorlog.ColoredFormatter(
            console_fmt,
            secondary_log_colors=secondary_log_colors,
        )
        console_handler.setFormatter(console_formatter)
        logger.addHandler(console_handler)
        return logger

    @staticmethod
    def _set_file_handler(
            logger: Logger,
            file_level: str | int, file_fmt: str,
            file_path: str, file_mode: str, file_max_bytes: int, file_backup_count: int, file_encoding: str,
    ) -> Logger:
        file_handler = _logger.RotatingFileHandler(
            file_path,
            mode=file_mode, maxBytes=file_max_bytes, backupCount=file_backup_count, encoding=file_encoding
        )
        file_handler.setLevel(file_level)
        file_formatter = logging.Formatter(file_fmt)
        file_handler.setFormatter(file_formatter)
        logger.addHandler(file_handler)
        return logger

    @staticmethod
    def get_logger(
            name: str | None = None,
            level: str = "DEBUG",
            to_console: bool = True,
            console_level: str = "DEBUG",
            console_fmt: str = (
                    "%(asctime_log_color)s%(asctime)s %(reset)s| "
                    "%(name_log_color)s%(name)s %(reset)s| "
                    "%(levelname_log_color)s%(levelname)s %(reset)s| "
                    # "%(process_log_color)sProcess: %(process)d %(processName_log_color)s(%(processName)s) %(reset)s| "
                    # "%(thread_log_color)sThread: %(thread)d %(threadName_log_color)s(%(threadName)s) %(reset)s| "
                    # "%(pathname_log_color)s%(pathname)s %(reset)s| "
                    "%(funcName_log_color)s%(funcName)s:%(lineno)d %(reset)s- "
                    "%(message_log_color)s%(message)s"
            ),
            to_file: bool = False,
            file_level: str = "DEBUG",
            file_path: str | None = None,
            file_mode: str = "a",
            file_max_bytes: int = 10 * 1024 * 1024,
            file_backup_count: int = 20,
            file_encoding: str = "utf8",
            file_fmt: str = (
                    "%(asctime)s | "
                    "%(name)s | "
                    "%(levelname)s | "
                    # "Process: %(process)d (%(processName)s) | "
                    # "Thread: %(thread)d (%(threadName)s) | "
                    # "%(pathname)s | "
                    "%(funcName)s:%(lineno)d - "
                    "%(message)s"
            ),
    ) -> Logger:
        _path = os.path.abspath(sys.argv[0])
        _name = os.path.basename(_path)
        _prefix = os.path.splitext(_name)[0]
        _file_name = _prefix + ".log"
        _file_dir = os.path.dirname(_path)
        _file_path = os.path.join(_file_dir, _file_name)

        if name is None:
            name = _prefix
        logger = cast(_logger.Logger, logging.getLogger(name))
        logger.setLevel(level)

        if to_console:
            console_handler_exists = any(isinstance(handler, logging.StreamHandler) for handler in logger.handlers)
            if not console_handler_exists:
                _logger._set_console_handler(
                    logger,
                    console_level=console_level,
                    console_fmt=console_fmt
                )
        else:
            for handler in logger.handlers:
                if isinstance(handler, logging.StreamHandler):
                    logger.removeHandler(handler)

        if to_file:
            file_handler_exists = any(isinstance(handler, logging.FileHandler) for handler in logger.handlers)
            if not file_handler_exists:
                if file_path is None:
                    file_path = _file_path
                _logger._set_file_handler(
                    logger,
                    file_level=file_level,
                    file_path=file_path, file_mode=file_mode, file_max_bytes=file_max_bytes,
                    file_backup_count=file_backup_count, file_encoding=file_encoding,
                    file_fmt=file_fmt
                )
        else:
            for handler in logger.handlers:
                if isinstance(handler, logging.FileHandler):
                    logger.removeHandler(handler)

        return logger

    logger: Logger | None = None


_logger.logger = _logger.get_logger()

__all__ = [
    "_logger"
]



import subprocess
from dataclasses import dataclass
from typing import Any

import execjs
import orjson
import py_mini_racer
from charset_normalizer import from_bytes

from utils._logger import _logger


class _execute:
    class js:
        @staticmethod
        def _get_js_code(
                js_code: str | None = None,
                js_file_path: str | None = None
        ) -> str:
            if js_code is None and js_file_path is None:
                raise ValueError(
                    f"Either js_code: {js_code!r} or js_file_path: {js_file_path!r} must be provided"
                )
            if js_code is not None:
                return js_code
            if js_file_path is not None:
                with open(js_file_path, "r", encoding="utf-8") as f:
                    js_code = f.read()
            return js_code

        @staticmethod
        def execute_javascript_by_execjs(
                js_code: str | None = None,
                js_file_path: str | None = None,
                func_name: str | None = None,
                func_args: tuple[Any, ...] | None = None
        ) -> Any:
            """
            # language=javascript
            js_code = '''
                  function sdk () {
                    let sum = 0;
                    for (const n of arguments) {
                      if (typeof n === "number") sum += n;
                    }
                    return sum;
                  }
                  '''
            result = _execute.js.execute_javascript_by_execjs(js_code, func_name="sdk", func_args=(1, 2, "3"))
            print(result)

            Args:
                js_code:
                js_file_path:
                func_name:
                func_args:

            Returns:

            """
            js_code = _execute.js._get_js_code(js_code, js_file_path)

            ctx = execjs.compile(js_code)
            if func_name is None:
                result = ctx.eval(js_code)
                return result
            if func_args is None:
                func_args = tuple()
            result = ctx.call(func_name, *func_args)
            return result

        @staticmethod
        def execute_javascript_by_py_mini_racer(
                js_code: str | None = None,
                js_file_path: str | None = None,
                func_name: str | None = None,
                func_args: tuple[Any, ...] | None = None
        ) -> Any:
            """
            # language=javascript
            js_code = '''
                  function sdk () {
                    let sum = 0;
                    for (const n of arguments) {
                      if (typeof n === "number") sum += n;
                    }
                    return sum;
                  }
                  '''
            result = _execute.js.execute_javascript_by_py_mini_racer(js_code, func_name="sdk", func_args=(1, 2, "3"))
            print(result)

            Args:
                js_code:
                js_file_path:
                func_name:
                func_args:

            Returns:

            """
            js_code = _execute.js._get_js_code(js_code, js_file_path)

            ctx = py_mini_racer.MiniRacer()
            result = ctx.eval(js_code)
            if func_name is None:
                return result
            if func_args is None:
                func_args = tuple()
            result = ctx.call(func_name, *func_args)
            return result

        @staticmethod
        def execute_javascript_by_subprocess(
                js_code: str | None = None,
                js_file_path: str | None = None,
                arguments: tuple[Any, ...] | None = None,
        ) -> Any:
            """
            # language=javascript
            js_code = '''(function () {
              arguments = process.argv.slice(1).map(JSON.parse);
              let sum = 0;
              for (const n of arguments) {
                if (typeof n === "number") sum += n;
              }
              console.log(JSON.stringify({ "sum": sum }));
            })();'''

            result = _execute.js.execute_javascript_by_subprocess(js_code, arguments=(1, 2, "3",))
            print(result["sum"])

            Args:
                js_code:
                js_file_path:
                arguments:

            Returns:

            """
            js_code = _execute.js._get_js_code(js_code, js_file_path)

            args = ["node", "-e", js_code] + list(map(lambda x: orjson.dumps(x).decode(), arguments))
            process = subprocess.Popen(
                args,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE
            )
            stdout, stderr = process.communicate()
            result = stdout.decode()
            result = orjson.loads(result)
            return result

    class cmd:
        @dataclass(frozen=True)
        class SubprocessPopenResult:
            returncode: int  # noqa
            stdout_content: bytes | None
            stdout_text: str
            stderr_content: bytes | None
            stderr_text: str

        @dataclass(frozen=True)
        class SubprocessRunResult:
            result: subprocess.CompletedProcess
            stdout_content: bytes
            stdout_text: str
            stderr_content: bytes
            stderr_text: str

        @staticmethod
        def _get_text(data: bytes, encoding: str | None = None) -> str:
            if encoding:
                text = data.decode(encoding)
            elif best_match := from_bytes(data).best():
                text = data.decode(best_match.encoding)
            else:
                text = str(data)
            return text

        @staticmethod
        def execute_cmd_code_by_subprocess_popen(
                cmd_code: str,
                encoding: str | None = None,
                logger: _logger.Logger | None = None
        ) -> SubprocessPopenResult:
            if logger:
                logger.debug(f'''$ {cmd_code}''')

            process = subprocess.Popen(
                cmd_code, shell=True,
                stdout=subprocess.PIPE, stderr=subprocess.PIPE
            )

            stdout_content, stdout_text = process.stdout, str()

            for i in process.stdout:
                text = _execute.cmd._get_text(i, encoding=encoding)
                stdout_text += text
                if text:
                    if logger:
                        logger.success("[√] " + text)

            stderr_content, stderr_text = process.stderr, str()
            for i in process.stderr:
                text = _execute.cmd._get_text(i, encoding=encoding)
                stderr_text += text
                if text:
                    if logger:
                        logger.error("[×] " + text)

            returncode = process.wait()

            return _execute.cmd.SubprocessPopenResult(
                returncode, stdout_content, stdout_text, stderr_content, stderr_text
            )

        @staticmethod
        def execute_cmd_code_by_subprocess_run(
                cmd_code: str,
                encoding: str | None = None,
                logger: _logger.Logger | None = None
        ) -> SubprocessRunResult:
            if logger:
                logger.debug(f'''> {cmd_code}''')

            result = subprocess.run(cmd_code, shell=True, capture_output=True)

            stdout_content, stdout_text = result.stdout, _execute.cmd._get_text(result.stdout, encoding=encoding)
            if stdout_text:
                if logger:
                    logger.success("[√] " + stdout_text)

            stderr_content, stderr_text = result.stderr, _execute.cmd._get_text(result.stderr, encoding=encoding)
            if stderr_text:
                if logger:
                    logger.error("[×] " + stderr_text)

            return _execute.cmd.SubprocessRunResult(
                result, stdout_content, stdout_text, stderr_content, stderr_text
            )


__all__ = [
    "_execute"
]



import inspect
import os


class _assets:
    @staticmethod
    def get_assets_file_path(file_name: str) -> str:
        _src_dir_path = os.path.splitext(os.path.abspath(__file__))[0]

        stack = inspect.stack()
        caller_frame = stack[1]
        _dest_dir_path = os.path.splitext(os.path.abspath(caller_frame.filename))[0]
        _rel_dir_path = os.path.relpath(_dest_dir_path, _src_dir_path).lstrip(os.path.pardir + os.path.sep)

        dir_path = os.path.join(_src_dir_path, _rel_dir_path)
        file_path = os.path.join(dir_path, file_name)
        return file_path


__all__ = [
    "_assets"
]



from .__string import _string as string
from ._assets import _assets as assets
from ._cookies import _cookies as cookies
from ._datetime import _datetime as datetime
from ._execute import _execute as execute
from ._file import _file as file
from ._headers import _headers as headers
from ._html import _html as html
from ._image import _image as image
from ._list import _list as list  # noqa
from ._logger import _logger as logger
from ._math import _math as math
from ._pandas import _pandas as pandas
from ._sql import _sql as sql
from ._types import _types as types
from ._url import _url as url
from ._validators import _validators as validators

__all__ = [
    "string",
    "assets",
    "cookies",
    "datetime",
    "execute",
    "file",
    "headers",
    "html",
    "image",
    "list",
    "logger",
    "math",
    "pandas",
    "sql",
    "types",
    "url",
    "validators",
]



import arrow


class _datetime:
    @staticmethod
    def now(fmt: str = "YYYY-MM-DD HH:mm:ss") -> str:
        return arrow.now().format(fmt)

    @staticmethod
    def delta(start_datetime_str: str, end_datetime_str: str) -> int:
        start_datetime = arrow.get(start_datetime_str).datetime
        end_datetime = arrow.get(end_datetime_str).datetime
        seconds = (end_datetime - start_datetime).total_seconds()
        return int(seconds)


__all__ = [
    "_datetime"
]



from typing import Any


class _cookies:
    @staticmethod
    def list_dict_to_str(cookies_list_dict: list[dict[str, Any]]) -> str:
        return "; ".join([i["name"] + "=" + i["value"] for i in cookies_list_dict])

    @staticmethod
    def str_to_dict(cookies_str: str) -> dict[str, str]:
        return {i.split("=")[0].strip(): i.split("=")[-1].strip() for i in cookies_str.split(";")}

    @staticmethod
    def dict_to_str(cookies_dict: dict[str, str]) -> str:
        return "; ".join([k + "=" + v for k, v in cookies_dict.items()])


__all__ = [
    "_cookies"
]
